﻿using System;
using itk.simple;

namespace DeformableRegistration
{
    class IterationUpdate : Command
    {

        private ImageRegistrationMethod m_Method;

        public IterationUpdate(ImageRegistrationMethod m)
        {
            m_Method = m;
        }

        public override void Execute()
        {
            VectorDouble pos = m_Method.GetOptimizerPosition();
            Console.WriteLine("{0} = {1} : [{2}, {3}, {4}, {5}, {6}, {7}]",
                              m_Method.GetOptimizerIteration(),
                              m_Method.GetMetricValue(),
                              pos[0], pos[1], pos[2], pos[3], pos[4], pos[5]);
        }
    }

    class Program
    {
        static void Main()
        {
            System.Diagnostics.Stopwatch watch = new System.Diagnostics.Stopwatch();
            watch.Start();

            string fixedImageFName = @"D:\Projects\ITK\testData\referencepet.mha";
            string movingImageFName = @"D:\Projects\ITK\testData\smoothpet.mha";
            string rigidTransformFName = @"D:\Projects\ITK\testData\RigidTransform.txt";

            ImageFileReader reader = new ImageFileReader();
            reader.SetOutputPixelType(PixelIDValueEnum.sitkFloat32);

            // Read the Fixed and Moving images.
            reader.SetFileName(fixedImageFName);
            Image fixedImage = reader.Execute();

            reader.SetFileName(movingImageFName);
            Image movingImage = reader.Execute();

            ImageRegistrationMethod R = new ImageRegistrationMethod();

            //set the interpolator
            R.SetInterpolator(InterpolatorEnum.sitkLinear);

            // Setup the metric parameters
            uint numberOfHistogramBins = 50;
            R.SetMetricAsMattesMutualInformation(numberOfHistogramBins);
            //zivy: explicitly set metric sampling strategy and percentage, the
            //options are random, regular, none. For the metric sampling percentage
            //set the random seed to a fixed value for reproducible results.
            R.SetMetricSamplingStrategy(R.RANDOM)
            R.SetMetricSamplingPercentage(0.01)

            //  Initialize a rigid transform by using Image Intensity Moments
            VectorDouble axis = new VectorDouble(new double[3]);

            axis[0] = 0.0;
            axis[1] = 0.0;
            axis[2] = 1.0;
            double angle = 0.0;
            VersorRigid3DTransform versorRigid3DTransform = new VersorRigid3DTransform(axis, angle);
            Transform rigidTransform = SimpleITK.CenteredTransformInitializer(fixedImage, movingImage, versorRigid3DTransform,
                                                                  CenteredTransformInitializerFilter.OperationModeType.MOMENTS);
            R.SetInitialTransform(rigidTransform);

            //zivy: user no longer needs to directly scale the parameters, the recommended
            //way is to use the automated method
            //  Define optimizer normaliztion to compensate for different dynamic range
            //  of rotations and translations.
            //VectorDouble optimimzerScales = new VectorDouble(new double[6]);
            //optimimzerScales[0] = 1.0;
            //optimimzerScales[1] = 1.0;
            //optimimzerScales[2] = 1.0;
            //double translationScale = 1.0 / 1000.0;
            //optimimzerScales[3] = translationScale;
            //optimimzerScales[4] = translationScale;
            //optimimzerScales[5] = translationScale;
            //R.SetOptimizerScales(optimimzerScales);
            R.SetOptimizerScalesFromPhysicalShift();

            // No ITK Equivalent to learningRate?
            double learningRate = 0.2;

            double minStep = 0.0001;
            uint numberOfIterations = 200;
            double relaxationFactor = 0.5;
            double maxStep = 0.2;
            double gradientMagnitudeTolerance = 0.0001;
            R.SetOptimizerAsRegularStepGradientDescent(learningRate, minStep, numberOfIterations, relaxationFactor, gradientMagnitudeTolerance,
                                                       ImageRegistrationMethod.EstimateLearningRateType.Never, maxStep);

            IterationUpdate cmd = new IterationUpdate(R);
            R.AddCommand(EventEnum.sitkIterationEvent, cmd);
            Console.WriteLine("Starting Rigid Registration");
            Transform outRigidTx = R.Execute(fixedImage, movingImage);
            string stopCondition = R.GetOptimizerStopConditionDescription();
            Console.WriteLine("Optimizer Stop Condition: {0}", stopCondition);
            Console.WriteLine("Rigid Registration Completed");

            watch.Stop();
            Console.WriteLine("duration: {0} seconds", (watch.ElapsedMilliseconds/1000.0).ToString());

            // Save the transform parameters in a file
            outRigidTx.WriteTransform(rigidTransformFName);

        }
    }
}
